"""
A module to load data from the standard directory structure (i.e. the one followed by csuite)
"""
import json
import logging
import os
from enum import Enum
from functools import partial
from typing import Any, Counter, Optional

import fsspec
import numpy as np
import torch
from tensordict import TensorDict

from causica.datasets.interventional_data import CounterfactualData, InterventionData
from causica.datasets.tensordict_utils import convert_one_hot
from causica.datasets.variable_types import DTYPE_MAP, VariableTypeEnum

CSUITE_DATASETS_PATH = "https://azuastoragepublic.blob.core.windows.net/datasets"


InterventionWithEffects = tuple[InterventionData, InterventionData, set[str]]
CounterfactualWithEffects = tuple[CounterfactualData, Optional[CounterfactualData], set[str]]

logger = logging.getLogger(__name__)


class DataEnum(Enum):
    TRAIN = "train.csv"
    TEST = "test.csv"
    INTERVENTIONS = "interventions.json"
    COUNTERFACTUALS = "counterfactuals.json"
    TRUE_ADJACENCY = "adj_matrix.csv"
    VARIABLES_JSON = "variables.json"


def convert_variable_types_to_enum(variable_metadata: dict[str, Any]):
    # Convert the types to the enum
    for variable in variable_metadata["variables"]:
        variable["type"] = VariableTypeEnum(variable["type"])

    return variable_metadata


def convert_enum_to_variable_types(variable_metadata: dict[str, Any]):
    # Convert enum to types
    for variable in variable_metadata["variables"]:
        variable["type"] = VariableTypeEnum(variable["type"]).value

    return variable_metadata


def load_data(
    root_path: str,
    data_enum: DataEnum,
    variables_metadata: Optional[dict[str, Any]] = None,
    **storage_options: dict[str, Any],
):
    """
    Load the Data from the location, dataset name and type of data.

    Args:
        root_path: The root path to the Data e.g. `CSUITE_DATASETS_PATH/csuite_linexp_2`
        data_enum: The type of dataset for which to return the path
        variables_metadata: Optional variables object (to save downloading it multiple times)
        **storage_options: Keyword args passed to `fsspec.open`
    Return:
        The downloaded data (the type depends on the data requested)
    """

    path_name = os.path.join(root_path, data_enum.value)

    fsspec_open = partial(fsspec.open, mode="r", encoding="utf-8", **storage_options)

    logger.debug(f"Loading {data_enum} from {path_name} with storage options {storage_options}")

    if data_enum == DataEnum.TRUE_ADJACENCY:
        with fsspec_open(path_name) as f:
            return torch.tensor(np.loadtxt(f, dtype=int, delimiter=","))

    if data_enum == DataEnum.VARIABLES_JSON:
        if variables_metadata is not None:
            raise ValueError("Variables metadata was supplied and requested")
        with fsspec_open(path_name) as f:
            return convert_variable_types_to_enum(json.load(f))

    if variables_metadata is None:
        variables_metadata = load_data(root_path, data_enum=DataEnum.VARIABLES_JSON)

    with fsspec_open(path_name) as f:
        if data_enum in {DataEnum.TRAIN, DataEnum.TEST}:
            arr = np.loadtxt(f, delimiter=",")
            categorical_sizes = _get_categorical_sizes(variables_list=variables_metadata["variables"])
            return convert_one_hot(
                tensordict_from_variables_metadata(arr, variables_metadata["variables"]),
                one_hot_sizes=categorical_sizes,
            )
        elif data_enum == DataEnum.INTERVENTIONS:
            return _load_interventions(json_object=json.load(f), metadata=variables_metadata)
        elif data_enum == DataEnum.COUNTERFACTUALS:
            return _load_counterfactuals(json_object=json.load(f), metadata=variables_metadata)
        else:
            raise RuntimeError("Unrecognized data type")


def _load_interventions(json_object: dict[str, Any], metadata: dict[str, Any]) -> list[InterventionWithEffects]:
    """
    Load the Interventional Datasets as a list of interventions/counterfactuals.

    Args:
        json_object: The .json file loaded as a json object.
        metadata: Metadata of the dataset containing names and types.

    Returns:
        A list of interventions and the nodes we want to observe for each
    """
    variables_list = metadata["variables"]

    intervened_column_to_group_name: dict[int, str] = dict(
        zip(
            json_object["metadata"]["columns_to_nodes"],
            list(item["group_name"] for item in variables_list),
        )
    )

    interventions_list = []
    for environment in json_object["environments"]:
        conditioning_idxs = environment["conditioning_idxs"]
        if conditioning_idxs is None:
            condition_nodes = []
        else:
            condition_nodes = [intervened_column_to_group_name[idx] for idx in environment["conditioning_idxs"]]

        intervention_nodes = [intervened_column_to_group_name[idx] for idx in environment["intervention_idxs"]]

        intervention = _to_intervention(
            np.array(environment["test_data"]),
            intervention_nodes=intervention_nodes,
            condition_nodes=condition_nodes,
            variables_list=variables_list,
        )
        # if the json has reference data create another intervention dataclass
        if (reference_data := environment["reference_data"]) is None:
            raise RuntimeError()
        reference = _to_intervention(
            np.array(reference_data),
            intervention_nodes=intervention_nodes,
            condition_nodes=condition_nodes,
            variables_list=variables_list,
        )

        # store the nodes we're interested in observing
        effect_nodes = set(intervened_column_to_group_name[idx] for idx in environment["effect_idxs"])
        # default to all nodes
        if not effect_nodes:
            effect_nodes = set(intervention.sampled_nodes)
        interventions_list.append((intervention, reference, effect_nodes))
    return interventions_list


def _load_counterfactuals(json_object: dict[str, Any], metadata: dict[str, Any]) -> list[CounterfactualWithEffects]:
    """
    Load the Interventional Datasets as a list of counterfactuals.

    Args:
        json_object: The .json file in loaded as a json object.
        metadata: Metadata of the dataset containing names and types.

    Returns:
        A list of counterfactuals and the nodes we want to observe for each
    """
    variables_list = metadata["variables"]

    intervened_column_to_group_name: dict[int, str] = dict(
        zip(
            json_object["metadata"]["columns_to_nodes"],
            list(item["group_name"] for item in variables_list),
        )
    )

    cf_list = []
    for environment in json_object["environments"]:
        factual_data = np.array(environment["conditioning_values"])
        intervention_nodes = [intervened_column_to_group_name[idx] for idx in environment["intervention_idxs"]]
        intervention = _to_counterfactual(
            np.array(environment["test_data"]),
            factual_data,
            intervention_nodes=intervention_nodes,
            variables_list=variables_list,
        )
        # if the json has reference data create another intervention dataclass
        if (reference_data := environment["reference_data"]) is None:
            reference = None
        else:
            reference = _to_counterfactual(
                np.array(reference_data),
                factual_data,
                intervention_nodes=intervention_nodes,
                variables_list=variables_list,
            )

        # store the nodes we're interested in observing
        effect_nodes = set(intervened_column_to_group_name[idx] for idx in environment["effect_idxs"])
        # default to all nodes
        if not effect_nodes:
            effect_nodes = set(intervention.sampled_nodes)
        cf_list.append((intervention, reference, effect_nodes))
    return cf_list


def _to_intervention(
    data: np.ndarray, intervention_nodes: list[str], condition_nodes: list[str], variables_list: list[dict[str, Any]]
) -> InterventionData:
    """Create an `InterventionData` object from the data within the json file."""
    interv_data = tensordict_from_variables_metadata(data, variables_list=variables_list)

    # all the intervention values in the dataset should be the same, so we use the first row
    first_row = interv_data[0]
    assert all(torch.allclose(interv_data[node_name], first_row[node_name]) for node_name in intervention_nodes)
    intervention_values = TensorDict(
        {node_name: first_row[node_name] for node_name in intervention_nodes}, batch_size=tuple()
    )

    condition_values = TensorDict(
        {node_name: first_row[node_name] for node_name in condition_nodes}, batch_size=tuple()
    )

    categorical_sizes = _get_categorical_sizes(variables_list=variables_list)

    return InterventionData(
        intervention_values=convert_one_hot(
            intervention_values, _intersect_dicts_left(categorical_sizes, intervention_values)
        ),
        intervention_data=convert_one_hot(interv_data, _intersect_dicts_left(categorical_sizes, interv_data)),
        condition_values=convert_one_hot(condition_values, _intersect_dicts_left(categorical_sizes, condition_values)),
    )


def _to_counterfactual(
    data: np.ndarray, base_data: np.ndarray, intervention_nodes: list[str], variables_list: list[dict[str, Any]]
) -> CounterfactualData:
    """Create an `CounterfactualData` object from the data within the json file."""
    interv_data = tensordict_from_variables_metadata(data, variables_list=variables_list)
    # all the intervention values in the dataset should be the same, so we use the first row
    first_row = interv_data[0]
    assert all(torch.allclose(interv_data[node_name], first_row[node_name]) for node_name in intervention_nodes)
    intervention_values = TensorDict(
        {node_name: first_row[node_name] for node_name in intervention_nodes}, batch_size=tuple()
    )
    factual_data = tensordict_from_variables_metadata(base_data, variables_list=variables_list)

    categorical_sizes = _get_categorical_sizes(variables_list=variables_list)

    return CounterfactualData(
        intervention_values=convert_one_hot(
            intervention_values, _intersect_dicts_left(categorical_sizes, intervention_values)
        ),
        counterfactual_data=convert_one_hot(interv_data, _intersect_dicts_left(categorical_sizes, interv_data)),
        factual_data=convert_one_hot(factual_data, _intersect_dicts_left(categorical_sizes, factual_data)),
    )


def _get_categorical_sizes(variables_list: list[dict[str, Any]]) -> dict[str, int]:
    categorical_sizes = {}
    for item in variables_list:
        if item["type"] == VariableTypeEnum.CATEGORICAL:
            upper = item.get("upper")
            lower = item.get("lower")
            if upper is not None and lower is not None:
                categorical_sizes[item["group_name"]] = item["upper"] - item["lower"] + 1
            else:
                assert upper is None and lower is None, "Please specify either both limits or neither"
                categorical_sizes[item["group_name"]] = -1
    return categorical_sizes


def tensordict_from_variables_metadata(data: np.ndarray, variables_list: list[dict[str, Any]]) -> TensorDict:
    """Returns a tensor created by concatenating all values along the last dim."""
    assert data.ndim == 2, "Numpy loading only supported for 2d data"
    batch_size = data.shape[0]

    # guaranteed to be ordered correctly in python 3.7+ https://docs.python.org/3/library/collections.html#collections.Counter
    sizes = Counter(d["group_name"] for d in variables_list)  # get the dimensions of each key from the variables
    assert sum(sizes.values()) == data.shape[1], "Variable sizes do not match data shape"

    # NOTE: This assumes that variables in the same group will have the same type.
    dtypes = {item["group_name"]: DTYPE_MAP[item["type"]] for item in variables_list}

    # slice the numpy array and assign the slices to the values of keys in the dictionary
    d = TensorDict({}, batch_size=batch_size)
    curr_idx = 0
    for key, length in sizes.items():
        d[key] = torch.Tensor(data[:, curr_idx : curr_idx + length]).to(dtype=dtypes[key])
        curr_idx += length

    return d


def tensordict_to_tensor(tensor_dict: TensorDict) -> torch.Tensor:
    """
    Convert a `TensorDict` into a 2D `torch.Tensor`.
    """
    return torch.cat(tuple(tensor_dict.values()), dim=-1)


def _intersect_dicts_left(dict_1: dict, dict_2: dict) -> dict:
    """Select the keys that are in both dictionaries, with values from the first."""
    return {key: dict_1[key] for key in dict_1.keys() & dict_2.keys()}



from causica.distributions.noise.bernoulli import BernoulliNoise, BernoulliNoiseModule
from causica.distributions.noise.categorical import CategoricalNoise, CategoricalNoiseModule
from causica.distributions.noise.joint import ContinuousNoiseDist, JointNoise, JointNoiseModule, create_noise_modules
from causica.distributions.noise.noise import IndependentNoise, Noise, NoiseModule
from causica.distributions.noise.spline import SplineNoise, SplineNoiseModule, create_spline_dist_params
from causica.distributions.noise.univariate_normal import UnivariateNormalNoise, UnivariateNormalNoiseModule



import logging
from typing import Any, Optional, Sequence, Union

import fsspec
import numpy as np
import pytorch_lightning as pl
import torch
from pytorch_lightning.trainer.states import TrainerFn
from pytorch_lightning.utilities.types import STEP_OUTPUT
from tensordict import TensorDict

from causica.datasets.causica_dataset_format import CounterfactualWithEffects, InterventionWithEffects
from causica.distributions import (
    AdjacencyDistribution,
    ConstrainedAdjacency,
    DistributionModule,
    ENCOAdjacencyDistributionModule,
    ExpertGraphContainer,
    GibbsDAGPrior,
    JointNoiseModule,
    SEMDistributionModule,
    create_noise_modules,
)
from causica.distributions.noise.joint import ContinuousNoiseDist
from causica.fsspec_helpers import get_storage_options_for_path
from causica.functional_relationships import ICGNN
from causica.graph.dag_constraint import calculate_dagness
from causica.graph.evaluation_metrics import adjacency_f1, orientation_f1
from causica.lightning.callbacks import AuglagLRCallback
from causica.lightning.data_modules.deci_data_module import DECIDataModule
from causica.lightning.modules.variable_spec_module import VariableSpecModule
from causica.sem.structural_equation_model import SEM
from causica.training.auglag import AugLagLossCalculator, AugLagLR, AugLagLRConfig
from causica.training.evaluation import (
    eval_ate_rmse,
    eval_intervention_likelihoods,
    eval_ite_rmse,
    list_logsumexp,
    list_mean,
)
from causica.training.training_callbacks import AverageMetricTracker

logging.basicConfig(level=logging.INFO)

NUM_GRAPH_SAMPLES = 100
NUM_ATE_ITE_SEMS = 10


class DECIModule(VariableSpecModule):
    """A PyTorch Lightning Module for running a DECI model."""

    def __init__(
        self,
        noise_dist: ContinuousNoiseDist = ContinuousNoiseDist.SPLINE,
        embedding_size: int = 32,
        out_dim_g: int = 32,
        norm_layer: bool = True,
        res_connection: bool = True,
        init_alpha: float = 0.0,
        init_rho: float = 1.0,
        prior_sparsity_lambda: float = 0.05,
        gumbel_temp: float = 0.25,
        auglag_config: Optional[AugLagLRConfig] = None,
        expert_graph_container: Optional[ExpertGraphContainer] = None,
        constraint_matrix_path: Optional[str] = None,
    ):
        super().__init__()
        self.auglag_config = auglag_config if auglag_config is not None else AugLagLRConfig()
        self.constraint_matrix_path = constraint_matrix_path
        self.constraint_matrix: Optional[torch.Tensor] = None

        self.embedding_size = embedding_size
        self.expert_graph_container: Optional[ExpertGraphContainer] = expert_graph_container

        self.gumbel_temp = gumbel_temp
        self.init_alpha = init_alpha
        self.init_rho = init_rho
        self.is_setup = False
        self.noise_dist = noise_dist
        self.norm_layer = norm_layer
        self.out_dim_g = out_dim_g
        self.prior_sparsity_lambda = prior_sparsity_lambda
        self.res_connection = res_connection

        # Inferred once the datamodule is available using `self.infer_missing_state_from_dataset()`
        self.dataset_name = None
        self.num_samples = None
        self.variable_group_shapes = None
        self.variable_types = None
        self.variable_names = None

    def prepare_data(self) -> None:
        """Set the constraint matrix (if necessary)."""
        if self.constraint_matrix_path:
            # NOTE: This assumes that adlfs paths will be opened non-anonymously
            storage_options = get_storage_options_for_path(self.constraint_matrix_path)

            with fsspec.open(self.constraint_matrix_path, **storage_options) as f:
                if self.constraint_matrix_path.endswith("npy"):
                    constraint_matrix = np.load(f)
                else:
                    raise ValueError(f"Unsupported constraint matrix format: {self.constraint_matrix_path}")
            self.constraint_matrix = torch.tensor(constraint_matrix)
        return super().prepare_data()

    def infer_missing_state_from_dataset(self):
        dataset_defined_members = [
            self.dataset_name,
            self.num_samples,
            self.variable_group_shapes,
            self.variable_types,
            self.variable_names,
        ]
        if any(member is None for member in dataset_defined_members):
            datamodule = getattr(self.trainer, "datamodule", None)
            if not isinstance(datamodule, DECIDataModule):
                raise TypeError(
                    f"Incompatible data module {datamodule}, requires a DECIDataModule but is "
                    f"{type(datamodule).mro()}"
                )
            if self.dataset_name is None:
                self.dataset_name = datamodule.dataset_name
            if self.num_samples is None:
                self.num_samples = len(datamodule.dataset_train)
            if self.variable_group_shapes is None:
                self.variable_group_shapes = datamodule.variable_shapes
            if self.variable_types is None:
                self.variable_types = datamodule.variable_types
            if self.variable_names is None:
                self.variable_names = datamodule.column_names

    def setup(self, stage: Optional[str] = None):
        if self.is_setup:
            return  # Already setup
        elif stage not in {TrainerFn.TESTING, TrainerFn.FITTING}:
            raise ValueError(f"Model can only be setup during the {TrainerFn.FITTING} and {TrainerFn.TESTING} stages.")

        self.infer_missing_state_from_dataset()
        assert self.dataset_name is not None
        assert self.num_samples is not None
        assert self.variable_group_shapes is not None
        assert self.variable_types is not None
        assert self.variable_names is not None

        self.node_names = list(self.variable_group_shapes)
        num_nodes = len(self.variable_group_shapes)

        adjacency_dist: DistributionModule[AdjacencyDistribution] = ENCOAdjacencyDistributionModule(num_nodes)
        if self.constraint_matrix is not None:
            adjacency_dist = ConstrainedAdjacency(adjacency_dist, self.constraint_matrix)

        icgnn = ICGNN(
            variables=self.variable_group_shapes,
            embedding_size=self.embedding_size,
            out_dim_g=self.out_dim_g,
            norm_layer=None if self.norm_layer is False else torch.nn.LayerNorm,
            res_connection=self.res_connection,
        )
        noise_submodules = create_noise_modules(self.variable_group_shapes, self.variable_types, self.noise_dist)
        noise_module = JointNoiseModule(noise_submodules)
        self.sem_module: SEMDistributionModule = SEMDistributionModule(adjacency_dist, icgnn, noise_module)
        self.auglag_loss: AugLagLossCalculator = AugLagLossCalculator(
            init_alpha=self.init_alpha, init_rho=self.init_rho
        )
        self.prior: GibbsDAGPrior = GibbsDAGPrior(
            num_nodes=num_nodes,
            sparsity_lambda=self.prior_sparsity_lambda,
            expert_graph_container=self.expert_graph_container,
        )
        # TODO: Set a more reasonable averaging period
        self.average_batch_log_prob_tracker: AverageMetricTracker = AverageMetricTracker(averaging_period=10)
        self.is_setup = True

    def training_step(self, *args, **kwargs) -> STEP_OUTPUT:
        _ = kwargs
        batch, *_ = args
        batch = batch.apply(lambda t: t.to(torch.float32))

        sem_distribution = self.sem_module()
        sem, *_ = sem_distribution.relaxed_sample(torch.Size([]), temperature=self.gumbel_temp)  # soft sample

        batch_log_prob = sem.log_prob(batch).mean()
        sem_distribution_entropy = sem_distribution.entropy()
        prior_term = self.prior.log_prob(sem.graph)
        objective = (-sem_distribution_entropy - prior_term) / self.num_samples - batch_log_prob
        constraint = calculate_dagness(sem.graph)
        self.average_batch_log_prob_tracker.step(batch_log_prob.item())
        step_output = {
            "loss": self.auglag_loss(objective, constraint / self.num_samples),
            "batch_log_prob": batch_log_prob,
            "constraint": constraint,
            "num_edges": (sem.graph > 0.0).count_nonzero(),
            "average_batch_log_prob": self.average_batch_log_prob_tracker.average,
            "vardist_entropy": sem_distribution_entropy,
            "prior_term": prior_term,
        }
        self.log_dict({"alpha": self.auglag_loss.alpha, "rho": self.auglag_loss.rho, **step_output}, prog_bar=True)
        return step_output

    def configure_optimizers(self):
        """Set the learning rates for different sets of parameters."""
        modules = {
            "icgnn": self.sem_module.functional_relationships,
            "vardist": self.sem_module.adjacency_module,
            "noise_dist": self.sem_module.noise_module,
        }
        parameter_list = [
            {"params": module.parameters(), "lr": self.auglag_config.lr_init_dict[name], "name": name}
            for name, module in modules.items()
        ]

        # Check that all modules are added to the parameter list
        check_modules = set(modules.values())
        for module in self.parameters(recurse=False):
            assert module in check_modules, f"Module {module} not in module list"

        return torch.optim.Adam(parameter_list)

    def configure_callbacks(self) -> Union[Sequence[pl.Callback], pl.Callback]:
        """Create a callback for the auglag callback."""
        return [AuglagLRCallback(AugLagLR(config=self.auglag_config))]

    def test_step_observational(self, batch: TensorDict, *args, **kwargs):
        """Evaluate the log prob of the model on the test set using multiple graph samples."""
        batch = batch.apply(lambda t: t.to(torch.float32))
        sems = self.sem_module().sample(torch.Size([NUM_GRAPH_SAMPLES]))
        dataset_size = self.trainer.datamodule.dataset_test.batch_size  # type: ignore
        assert len(dataset_size) == 1, "Only one batch size is supported"
        # estimate log prob for each sample using graph samples and report the mean over graphs
        log_prob_test = list_logsumexp([sem.log_prob(batch) for sem in sems]) - np.log(NUM_GRAPH_SAMPLES)
        self.log(
            "eval/test_LL",
            torch.sum(log_prob_test, dim=-1).item() / dataset_size[0],
            reduce_fx=sum,
            add_dataloader_idx=False,
        )

    def test_step_graph(self, true_adj_matrix: torch.Tensor, *args, **kwargs):
        """Evaluate the graph reconstruction performance of the model against the true graph."""
        _, _ = args, kwargs
        sem_samples = self.sem_module().sample(torch.Size([NUM_GRAPH_SAMPLES]))
        graph_samples = [sem.graph for sem in sem_samples]

        adj_f1 = list_mean([adjacency_f1(true_adj_matrix, graph) for graph in graph_samples]).item()
        orient_f1 = list_mean([orientation_f1(true_adj_matrix, graph) for graph in graph_samples]).item()
        self.log("eval/adjacency.f1", adj_f1, add_dataloader_idx=False)
        self.log("eval/orientation.f1", orient_f1, add_dataloader_idx=False)

    def test_step_interventions(self, interventions: InterventionWithEffects, *args, **kwargs):
        """Evaluate the ATE and Interventional log prob performance of the model"""
        _, _ = args, kwargs
        sems_list: list[SEM] = list(self.sem_module().sample(torch.Size([NUM_ATE_ITE_SEMS])))
        interventional_log_prob = eval_intervention_likelihoods(sems_list, interventions)
        self.log("eval/Interventional_LL", torch.mean(interventional_log_prob).item(), add_dataloader_idx=False)

        mean_ate_rmse = eval_ate_rmse(sems_list, interventions)
        self.log("eval/ATE_RMSE", list_mean(list(mean_ate_rmse.values())).item(), add_dataloader_idx=False)

    def test_step_counterfactuals(self, counterfactuals: CounterfactualWithEffects, *args, **kwargs):
        """Evaluate the ITE performance of the model"""
        _, _ = args, kwargs
        sems_list = list(self.sem_module().sample(torch.Size([NUM_ATE_ITE_SEMS])))
        ite_rmse = eval_ite_rmse(sems_list, counterfactuals)
        self.log("eval/ITE_RMSE", list_mean(list(ite_rmse.values())).item(), add_dataloader_idx=False)

    def on_save_checkpoint(self, checkpoint: dict[str, Any]) -> None:
        checkpoint["sem_module"] = self.sem_module



import time
from queue import Queue

import mlflow


class MLFlowBatch:
    """
    A class for collecting MLFlow logs and submitting them as a batch to speed up logging.

    It will automatically batch all logs, once the batch size is reached.

    Accumulated metrics can be manually logged by calling flush.
    """

    def __init__(self, batch_size: int) -> None:
        self._queue: Queue = Queue(maxsize=batch_size)  # mypy requires explicit type annotation here.
        self._client = mlflow.tracking.MlflowClient()
        active_run = mlflow.active_run()
        if active_run is None:
            raise ValueError("No active ML flow run.")
        self._run_id = active_run.info.run_id

    def log_metric(self, key: str, value: float, step: int = 0):
        """Stores a metric in the log batch and pushes if the batch is full."""
        metric = mlflow.entities.Metric(key=key, value=float(value), timestamp=int(time.time()), step=step)
        self._queue.put(metric)
        if self._queue.full():
            self.flush()

    def flush(self):
        """Manually push the accumulated logs."""
        metrics = []
        while not self._queue.empty():
            metrics.append(self._queue.get())
        self._client.log_batch(run_id=self._run_id, metrics=metrics)



import torch
from tensordict import TensorDict

from causica.functional_relationships.functional_relationships import FunctionalRelationships


class DoFunctionalRelationships(FunctionalRelationships):
    """
    A `FunctionalRelationship` that one can "do", i.e. condition nodes and cut the links to their parents.
    """

    def __init__(self, func: FunctionalRelationships, do: TensorDict, submatrix: torch.Tensor) -> None:
        """
        Args:
            func: The unintervened functional relationships
            do: the nodes on which to intervene
            submatrix: the submatrix that the unintervened nodes represent in the larger graph
        """
        assert all(val.ndim == 1 for val in do.values()), "Intervention is only supported for 1 vector per variable"

        new_variables = {key: value for key, value in func.variables.items() if key not in do.keys()}
        super().__init__(new_variables)

        self.func = func
        self.do = do  # dict of key to vectors
        self.submatrix = submatrix

        self.do_nodes_mask = torch.tensor(
            [(name in self.do.keys()) for name in self.func.variables.keys()], dtype=torch.bool
        )

    def pad_intervened_graphs(self, graphs: torch.Tensor) -> torch.Tensor:
        """
        Pad the intervened graph with the unintervened nodes.

        Args:
            graphs: Weighted adjacency matrix, size batch_size_g + (do_func_n, do_func_n)
        Returns:
            A tensor of shape batch_shape_g + (func_n, func_n)
        """
        target_shape = graphs.shape[:-2] + (self.func.num_nodes, self.func.num_nodes)

        output_graphs = torch.zeros(target_shape, dtype=graphs.dtype, device=graphs.device)
        assign_submatrix(output_graphs, graphs, ~self.do_nodes_mask, ~self.do_nodes_mask)
        assign_submatrix(output_graphs, self.submatrix, self.do_nodes_mask, ~self.do_nodes_mask)
        return output_graphs

    def forward(self, samples: TensorDict, graphs: torch.Tensor) -> TensorDict:
        """
        Args:
            samples: Batched inputs, size batch_size_x + (processed_dim_all).
            graphs: Weighted adjacency matrix, size batch_size_g + (n, n)
        Returns:
            A tensor of shape batch_shape_x + batch_shape_g + (processed_dim_all)
        """
        # add the expanded intervention values to the samples
        samples_with_do = samples.update(self.do.expand(*samples.batch_size))

        # create the full graphs
        graphs = self.pad_intervened_graphs(graphs)

        forward = self.func.forward(samples_with_do, graphs)
        return forward.select(*(key for key in forward.keys() if key not in self.do.keys()), inplace=True)


def assign_submatrix(A: torch.Tensor, B: torch.Tensor, x_mask: torch.Tensor, y_mask: torch.Tensor) -> None:
    """
    Assign `B` to a submatrix of `A`. The matrix `A` is changed in place.
    Args:
        A: tensor with >=2 dimensions `[*b, x, y]`
        B: tensor with dimensions `[*b, x', y']` where `x'<=x` and `y'<=y`
        x_mask: boolean tensor of shape `[*b, x]` indicating the rows of A that are to be updated. The sum
            of x_mask should equal `x'`
        y_mask: boolean tensor of shape `[*b, y]` indicating the columns of A that are to be updated. The sum
            of y_mask should equal `y'`
    """
    assign_mask = torch.ones_like(A, dtype=torch.bool)
    assign_mask[..., ~x_mask, :] = 0
    assign_mask[..., :, ~y_mask] = 0
    A[assign_mask] = B.flatten()


def create_do_functional_relationship(
    interventions: TensorDict, func: FunctionalRelationships, graph: torch.Tensor
) -> tuple[DoFunctionalRelationships, torch.Tensor]:
    """
    Given a set of interventions, `FunctionalRelationships` and a graph, create a `DoFunctionalRelationships` and an intervened graph.

    Args:
        interventions: the nodes and their intervention values
        func: the functional relationship of the unintervened SEM
        graph: the unintervened graph shape: [..., num_nodes, num_nodes]
    Return:
        A tuple with the intervened functional relationship and the intervened graph
    """
    node_names = list(func.variables.keys())
    do_nodes_mask = torch.zeros(len(node_names), dtype=torch.bool)
    for i, name in enumerate(node_names):
        if name in interventions.keys():
            do_nodes_mask[i] = 1

    do_graph = graph[..., ~do_nodes_mask, :][..., :, ~do_nodes_mask]
    submatrix = graph[..., do_nodes_mask, :][..., :, ~do_nodes_mask]
    return DoFunctionalRelationships(func, interventions, submatrix), do_graph



import logging

from causica.lightning.callbacks import MLFlowSaveConfigCallback
from causica.lightning.cli import LightningCLIWithDefaults
from causica.lightning.data_modules.deci_data_module import DECIDataModule
from causica.lightning.modules.deci_module import DECIModule

if __name__ == "__main__":
    # Set Azure logging to warning to prevent spam from HTTP requests
    logging.getLogger("azure").setLevel(logging.WARNING)

    cli = LightningCLIWithDefaults(
        DECIModule,
        DECIDataModule,
        run=False,
        save_config_callback=MLFlowSaveConfigCallback,
        subclass_mode_model=True,
        subclass_mode_data=True,
    )
    cli.trainer.fit(cli.model, datamodule=cli.datamodule)
    cli.trainer.test(cli.model, datamodule=cli.datamodule)



from typing import Optional, Type

import torch
from tensordict import TensorDict
from torch import nn

from causica.functional_relationships.functional_relationships import (
    FunctionalRelationships,
    sample_dict_to_tensor,
    tensor_to_sample_dict,
)


class ICGNN(FunctionalRelationships):
    """
    This is a `FunctionalRelationsips` that implements the ICGNN.

    Details can be found here: https://openreview.net/forum?id=S2pNPZM-w-f

    This wraps the `FGNNI` in a `TensorDict` interface.
    """

    def __init__(
        self,
        variables: dict[str, torch.Size],
        embedding_size: Optional[int] = None,
        out_dim_g: Optional[int] = None,
        norm_layer: Optional[Type[nn.LayerNorm]] = None,
        res_connection: bool = False,
    ) -> None:
        super().__init__(variables)

        # this needs to be a parameter so it is registered to the module
        self.stacked_variable_masks = torch.nn.Parameter(
            torch.stack(list(self.variable_masks.values())).float(), requires_grad=False
        )

        self.nn = FGNNI(self.stacked_variable_masks, embedding_size, out_dim_g, norm_layer, res_connection)

    def forward(self, samples: TensorDict, graphs: torch.Tensor) -> TensorDict:
        return tensor_to_sample_dict(
            self.nn(sample_dict_to_tensor(samples, self.variable_masks), graphs), self.variable_masks
        )


class FGNNI(nn.Module):
    """
    Defines the function f for the SEM. For each variable x_i we use
    f_i(x) = f(e_i, sum_{k in pa(i)} g(e_k, x_k)), where e_i is a learned embedding
    for node i.
    """

    def __init__(
        self,
        group_mask: torch.Tensor,
        embedding_size: Optional[int] = None,
        out_dim_g: Optional[int] = None,
        norm_layer: Optional[Type[nn.LayerNorm]] = None,
        res_connection: bool = False,
    ):
        """
        Args:
            group_mask: A mask of shape (num_nodes, num_processed_cols) such that group_mask[i, j] = 1. when col j is in group i.
            embedding_size: Size of the embeddings used by each node. If none, default is processed_dim_all.
            out_dim_g: Output dimension of the "inner" NN, g. If none, default is embedding size.
            layers_g: Size of the layers of NN g. Does not include input not output dim. If none, default
                      is [a], with a = max(2 * input_dim, embedding_size, 10).
            layers_f: Size of the layers of NN f. Does not include input nor output dim. If none, default
                      is [a], with a = max(2 * input_dim, embedding_size, 10)
        """
        super().__init__()
        self.group_mask = group_mask
        self.num_nodes, self.processed_dim_all = group_mask.shape
        # Initialize embeddings
        self.embedding_size = self.processed_dim_all if embedding_size is None else embedding_size
        aux = torch.randn(self.num_nodes, self.embedding_size) * 0.01
        self.embeddings = nn.parameter.Parameter(aux, requires_grad=True)  # Shape (input_dim, embedding_size)

        # Set value for out_dim_g
        out_dim_g = self.embedding_size if out_dim_g is None else out_dim_g
        # Set NNs sizes
        a = max(4 * self.processed_dim_all, self.embedding_size, 64)
        in_dim_g = self.embedding_size + self.processed_dim_all
        in_dim_f = self.embedding_size + out_dim_g
        self.g = generate_fully_connected(
            input_dim=in_dim_g,
            output_dim=out_dim_g,
            hidden_dims=[a, a],
            non_linearity=nn.LeakyReLU,
            activation=nn.Identity,
            device=group_mask.device,
            normalization=norm_layer,
            res_connection=res_connection,
        )
        self.f = generate_fully_connected(
            input_dim=in_dim_f,
            output_dim=self.processed_dim_all,
            hidden_dims=[a, a],
            non_linearity=nn.LeakyReLU,
            activation=nn.Identity,
            device=group_mask.device,
            normalization=norm_layer,
            res_connection=res_connection,
        )
        self.w = torch.nn.Parameter(torch.zeros((self.num_nodes, self.num_nodes)), requires_grad=True)

    def forward(self, samples: torch.Tensor, graphs: torch.Tensor) -> torch.Tensor:
        """
        Computes non-linear function h(X, W) using the given weighted adjacency matrix.

        g takes inputs of size batch_shape + (embedding_size + processed_dim_all) and outputs batch_shape + (out_dim_g)
        the input will be appropriately masked to correspond to one variable group

        f takes inputs of size batch_shape + (embedding_size + out_dim_g) and outputs batch_shape + (processed_dim_all)
        the ouptut is then masked to correspond to one variable

        Args:
            samples: Batched inputs, size batch_size_x + (processed_dim_all).
            graphs: Weighted adjacency matrix, size batch_size_g + (n, n)
        Returns:
            A tensor of shape batch_shape_x + batch_shape_g + (processed_dim_all)
        """
        batch_shape_x = samples.shape[:-1]
        batch_shape_g = graphs.shape[:-2]

        # Generate required input for g (concatenate X and embeddings)
        # Pointwise multiply X
        # Shape batch_shape_x + (num_nodes, processed_dim_all)
        masked_samples = torch.einsum("...i,ji->...ji", samples, self.group_mask)
        E = self.embeddings.expand(*batch_shape_x, -1, -1)  # Shape batch_shape_x + (num_nodes, embedding_size)

        # Shape batch_shape_x + (num_nodes, embedding_size + processed_dim_all)
        embedded_samples = self.g(
            torch.cat([masked_samples, E], dim=-1)
        )  # Shape batch_shape_x + (num_nodes, out_dim_g)

        target_shape = batch_shape_x + batch_shape_g + embedded_samples.shape[-2:]
        view_shape = batch_shape_x + (1,) * len(batch_shape_g) + embedded_samples.shape[-2:]
        # Shape batch_shape_x + batch_shape_g + (num_nodes, out_dim_g)
        embedded_samples_broad = embedded_samples.view(view_shape).expand(target_shape)
        # Aggregate sum and generate input for f (concatenate X_aggr and embeddings)
        # Shape batch_shape_x + batch_shape_g + (num_nodes, out_dim_g)
        samples_aggr_sum = torch.einsum("...jk,...jl->...lk", embedded_samples_broad, graphs * self.w)

        # expand dimensions of E batch_shape_x + batch_shape_g + (num_nodes, embedding_size)
        E_broad = E.view(view_shape).expand(target_shape)
        # Run f Shape batch_shape_x + batch_shape_g + (num_nodes, processed_dim_all)
        samples_rec = self.f(torch.cat([samples_aggr_sum, E_broad], dim=-1))
        # Mask and aggregate Shape batch_shape_x + batch_shape_g + (processed_dim_all)
        return torch.einsum("...ij,ij->...j", samples_rec, self.group_mask)


def generate_fully_connected(
    input_dim: int,
    output_dim: int,
    hidden_dims: list[int],
    non_linearity: Optional[Type[nn.Module]],
    activation: Optional[Type[nn.Module]],
    device: torch.device,
    p_dropout: float = 0.0,
    normalization: Optional[Type[nn.LayerNorm]] = None,
    res_connection: bool = False,
) -> nn.Module:
    """
    Generate a fully connected network.

    Args:
        input_dim: Int. Size of input to network.
        output_dim: Int. Size of output of network.
        hidden_dims: List of int. Sizes of internal hidden layers. i.e. [a, b] is three linear layers with shapes (input_dim, a), (a, b), (b, output_dim)
        non_linearity: Non linear activation function used between Linear layers.
        activation: Final layer activation to use.
        device: torch device to load weights to.
        p_dropout: Float. Dropout probability at the hidden layers.
        init_method: initialization method
        normalization: Normalisation layer to use (batchnorm, layer norm, etc). Will be placed before linear layers, excluding the input layer.
        res_connection : Whether to use residual connections where possible (if previous layer width matches next layer width)

    Returns:
        Sequential object containing the desired network.
    """
    layers: list[nn.Module] = []

    prev_dim = input_dim
    for idx, hidden_dim in enumerate(hidden_dims):

        block: list[nn.Module] = []

        if normalization is not None and idx > 0:
            block.append(normalization(prev_dim).to(device))
        block.append(nn.Linear(prev_dim, hidden_dim).to(device))

        if non_linearity is not None:
            block.append(non_linearity())
        if p_dropout != 0:
            block.append(nn.Dropout(p_dropout))

        if res_connection and (prev_dim == hidden_dim):
            layers.append(_ResBlock(nn.Sequential(*block)))
        else:
            layers.append(nn.Sequential(*block))
        prev_dim = hidden_dim

    if normalization is not None:
        layers.append(normalization(prev_dim).to(device))
    layers.append(nn.Linear(prev_dim, output_dim).to(device))

    if activation is not None:
        layers.append(activation())

    return nn.Sequential(*layers)


class _ResBlock(nn.Module):
    """
    Wraps an nn.Module, adding a skip connection to it.
    """

    def __init__(self, block: nn.Module):
        """
        Args:
            block: module to which skip connection will be added. The input dimension must match the output dimension.
        """
        super().__init__()
        self.block = block

    def forward(self, x):
        return x + self.block(x)
